#!/usr/bin/env python
# coding: utf-8

# Some L-systems:
#   https://www.robertdickau.com/lsys2d.html
#   https://resources.wolframcloud.com/FunctionRepository/resources/LSystemPlot/

# In[12]:
import math

def get_l_system_paths(depth, axiom, ignore, rules, dtheta,
        initial_theta, length_scale):
    '''Apply an L-system.'''
    # Build the final script.
    script = axiom
    for i in range(depth):
        new_script = ''
        for ch in script:
            if ch in rules:
                new_script += rules[ch]
            else:
                new_script += ch
        script = new_script
    # print(script)

    # Create the points.
    p1 = (0, 0)
    length = 100
    paths = []      # A list of paths
    path = [p1]     # The current path
    theta = initial_theta
    waypoints = []  # Places we will go back to
    wayangles = []  # Angles for the waypoints
    waylengths = []
    for ch in script:
        if ch in ignore:
            pass
        elif ch == '-':
            theta -= dtheta
        elif ch == '+':
            theta += dtheta
        elif ch == '[':
            waypoints.append(path[-1])
            wayangles.append(theta)
            waylengths.append(length)
        elif ch == ']':
            if len(path) > 1: paths.append(path)
            p1 = waypoints.pop()
            path = [p1]
            theta = wayangles.pop()
            length = waylengths.pop()
        else:
            x = p1[0] + length * math.cos(theta)
            y = p1[1] + length * math.sin(theta)
            p1 = (x, y)
            path.append(p1)
            length *=  length_scale

    if len(path) > 1:
        paths.append(path)

    # Return the points.
    return paths

#%%

def draw_point(canvas, pt, radius, outline, fill):
    canvas.create_oval(
        pt[0] - radius,
        pt[1] - radius,
        pt[0] + radius,
        pt[1] + radius,
        outline=outline,
        fill=fill)

def fit_curves_to_canvas(canvas, curves, margin):
    '''Scale a set of curves to fit a canvas.'''
    # Get the canvas's dimensions.
    canvas_wid = canvas.winfo_width() - 2 * margin
    canvas_hgt = canvas.winfo_height() - 2 * margin

    # Get the curves' dimensions.
    x1, y1, x2, y2 = curves_bbox(canvas, curves)
    curve_wid = x2 - x1
    curve_hgt = y2 - y1

    # Calculate the scale factor.
    x_scale = (canvas_wid - 2 * margin) / curve_wid
    y_scale = (canvas_hgt - 2 * margin) / curve_hgt
    scale = min(x_scale, y_scale)

    # Scale the curve.
    for curve in curves:
        canvas.scale(curve, 0, 0, scale, scale)

    # Center the curves.
    x1, y1, x2, y2 = curves_bbox(canvas, curves)
    cx = (x1 + x2) / 2
    cy = (y1 + y2) / 2
    xmin = cx - canvas_wid / 2
    xmax = cx + canvas_wid / 2
    ymin = cy - canvas_hgt / 2
    ymax = cy + canvas_hgt / 2
    canvas.config(scrollregion=(xmin, ymin, xmax, ymax))

def curves_bbox(canvas, curves):
    '''Return a bounding box for a set of curves.'''
    x1, y1, x2, y2 = canvas.bbox(curves[0])
    for curve in curves:
        new_x1, new_y1, new_x2, new_y2 = canvas.bbox(curve)
        if new_x1 < x1: x1 = new_x1
        if new_x2 > x2: x2 = new_x2
        if new_y1 < y1: y1 = new_y1
        if new_y2 > y2: y2 = new_y2
    return x1, y1, x2, y2


#%%
import tkinter as tk
from tkinter import messagebox

# Method for getting ints and floats from Entry widgets.'''

def make_field(parent, label_text, label_width, entry_text, entry_width):
    frame = tk.Frame(parent)
    frame.pack(side=tk.TOP, anchor=tk.W, padx=1, pady=3)
    label = tk.Label(frame, text=label_text, width=label_width, anchor=tk.W)
    label.pack(side=tk.LEFT)

    var = tk.StringVar(value=entry_text)
    entry = tk.Entry(frame, width=entry_width)
    entry.pack(side=tk.LEFT)
    return entry, var

def make_text(parent, label_text, label_width, text_width, text_height,
        text_text):
    frame = tk.Frame(parent, bg='white')
    frame.pack(side=tk.TOP, anchor=tk.W, padx=1, pady=3)
    label = tk.Label(frame, text=label_text, width=label_width, anchor=tk.NW)
    label.pack(side=tk.LEFT)

    text = tk.Text(frame, width=text_width, height=text_height)
    text.pack(side=tk.LEFT)
    text.insert(tk.END, text_text)
    return text

def get_int(name, var, min_value, max_value):
    try:
        value = int(var.get())
        if value >= min_value and value <= max_value:
            return value
    except:
        pass

    messagebox.showerror(f'{name} Error',
        f'{name} must be an integer between {min_value} and {max_value}.')
    return None

def get_float(name, var, min_value, max_value):
    try:
        value = float(var.get())
        if value >= min_value and value <= max_value:
            return value
    except:
        pass

    messagebox.showerror(f'{name} Error',
        f'{name} must be a number between {min_value} and {max_value}.')
    return None

# In[18]:
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

class LSystemsApp:
    def __init__(self):
        self.window = tk.Tk()
        self.window.title('l_systems_interactive')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('600x400')

        # Build the UI.
        l_frame = tk.Frame(self.window)
        l_frame.pack(side=tk.LEFT, anchor=tk.NW, padx=3, pady=3)

        # Each L-system includes:
        #   Name: value
        # Where value includes:
        #   Dictionary of rules
        #   Axiom
        #   Ignore
        #   Dtheta
        #   Max Depth
        #   Theta 0
        #   Length Scale
        self.lsystems = {
            'Spiral': (
                {'X': '-XX'},
                'X', '', -2, 20, 0, 0.995, 15),
            'Fern 3': (
                {'X': 'F+[[X]-X]-F[-FX-FX]+X', 'F': 'FF'},
                'X', 'X', -25, 9, -50, 1, 6),
            'Fern 2': (
                {'X': 'F+[[X]-X-X]-F[-FX]+X', 'F': 'FF'},
                'X', 'X', -30, 9, -50, 1, 5),
            '5-Tree': (
                {'F': 'F[-F][--F][+F][++F][F]'},
                'F', '', 30, 10, -90, 0.75, 5),
            '90 Degree Tree': (
                {'F': 'F[-F][+F]'},
                'F', '', 90, 10, -90, 0.75, 6),
            '60 Degree Tree': (
                {'F': 'F[-F][+F]'},
                'F', '', 60, 10, -90, 0.75, 10),
            '30 Degree Tree': (
                {'F': 'F[-F][+F]'},
                'F', '', 30, 10, -90, 0.75, 9),
            'Square-Tree': (
                {'F': 'FF[-F+F][+F-F]'},
                'F', '', 90, 10, -90, 0.95, 6),
            '3-Tree': (
                {'F': 'F[-F][F][+F]'},
                'F', '', 45, 10, -90, 0.75, 7),
            'Tree': (
                {'F': 'F[-F][+F]'},
                'F', '', 45, 10, -90, 0.75, 10),
            # https://en.wikipedia.org/wiki/L-system#Example_7:_fractal_plant
            # The original version uses axiom X and angle 25. I've modified
            # it to give it a proper orientation in our coordinate system.
            'Fern': (
                {'X': 'F+[[X]-X]-F[-FX]+X', 'F': 'FF'},
                'X', 'X', -25, 9, -50, 1, 7),
            '32-Segment Curve': (
                {'F': '-F+F-F-F+F+FF-F+F+FF+F-F-FF+FF-FF+F+F-FF-F-F+FF-F-F+F+F-F+'},
                'F+F+F+F', '', 90, 4, 0, 1, 2),
            'Box Fractal': (
                {'F': 'F-F+F+F-F'},
                'F-F-F-F', '', 90, 7, 0, 1, 3),
            'Dragon Curve': (
                {'X': 'X+YF+', 'Y': '-FX-Y'},
                'FX', '', 90, 20, 0, 1, 11),
            # Hilbert Curve: https://www.fractal.garden/l-system/hilbert-curve
            'Hilbert Curve': (
                {'V': '-WF+VFV+FW-', 'W': '+VF-WFW-FV+'},
                'W', 'VW', 90, 10, 0, 1, 5),
            'Hilbert Curve II': (
                {'X': 'XFYFX+F+YFXFY-F-XFYFX', 'Y': 'YFXFY-F-XFYFX+F+YFXFY'},
                'X', '', 90, 6, 0, 1, 3),
            'Koch Snowflake': (
                {'F': 'F+F--F+F'},
                'F--F--F', '', 60, 7, 0, 1, 4),
            'Peano Curve': (
                {'F': 'F+F-F-F-F+F+F+F-F'},
                'F', '', 90, 6, 0, 1, 3),
            'Peano-Gosper Curve': (
                {'X': 'X+YF++YF-FX--FXFX-YF+', 'Y': '-FX+YFYF++YF+FX--FX-Y'},
                'FX', '', 60, 7, 0, 1, 3),
            'Gosper Boundary': (
                {'A': 'A-A+A'},
                'A-A-A-A-A-A-', '', 60, 7, 0, 1, 4),
            'Quadratic Koch Island': (
                {'F': 'F-F+F+FFF-F-F+F'},
                'F+F+F+F', '', 90, 6, 0, 1, 2),
            'Sierpiński Arrowhead': (
                {'X': 'YF+XF+Y', 'Y': 'XF-YF-X'},
                'YF', '', 60, 10, 0, 1, 6),
            # Sierpiński Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Curve': (
                {'X': 'XF+G+XF--F--XF+G+X'},
                'F--XF--F--XF', 'X', 45, 9, 0, 1, 4),
            # Sierpiński Square Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Square Curve': (
                {'X': 'XF-F+F-XF+F+XF-F+F-X'},
                'F+XF+F+XF', '', 90, 9, 0, 1, 4),
            'Siepiński Sieve': (
                {'F': 'FF', 'X': '--FXF++FXF++FXF--'},
                'FXF--FF--FF', '', 60, 10, 0, 1, 4),
        }

        row = 0
        curve_names = list(self.lsystems.keys())
        self.curve_combo = ttk.Combobox(l_frame, values=curve_names,
            state='readonly', width=23)
        self.curve_combo.grid(row=row, column=0, columnspan=2, pady=(0, 20))
        self.curve_combo.current(0)
        row += 1

        label = tk.Label(l_frame, text='Axiom:')
        label.grid(row=row, column=0, padx=(0,10), sticky=tk.W)
        self.axiom_var = tk.StringVar(value='F')
        axiom_entry = tk.Entry(l_frame, textvariable=self.axiom_var, width=15)
        axiom_entry.grid(row=row, column=1, sticky=tk.W)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Ignore:')
        label.grid(row=row, column=0, padx=(0,10), sticky=tk.W)
        self.ignore_var = tk.StringVar(value='')
        ignore_entry = tk.Entry(l_frame, textvariable=self.ignore_var, width=15)
        ignore_entry.grid(row=row, column=1, sticky=tk.W)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Rules:')
        label.grid(row=row, column=0, sticky=tk.NW)
        row += 1
        self.rules_text = tk.Text(l_frame, width=19, height=5)
        self.rules_text.grid(row=row, column=0, columnspan=2,
            sticky=tk.NW, padx=5)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Dtheta:')
        label.grid(row=row, column=0, padx=(0,10), sticky=tk.W)
        self.dtheta_var = tk.StringVar(value=3)
        dtheta_entry = tk.Entry(l_frame, textvariable=self.dtheta_var, width=5,
            justify=tk.RIGHT)
        dtheta_entry.grid(row=row, column=1, sticky=tk.W)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Theta 0:')
        label.grid(row=row, column=0, padx=(0,10), sticky=tk.W)
        self.theta0_var = tk.StringVar(value=3)
        theta0_entry = tk.Entry(l_frame, textvariable=self.theta0_var, width=5,
            justify=tk.RIGHT)
        theta0_entry.grid(row=row, column=1, sticky=tk.W)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Scale:')
        label.grid(row=row, column=0, padx=(0,10), sticky=tk.W)
        self.scale_var = tk.StringVar(value=1)
        scale_entry = tk.Entry(l_frame, textvariable=self.scale_var, width=5,
            justify=tk.RIGHT)
        scale_entry.grid(row=row, column=1, sticky=tk.W)
        l_frame.rowconfigure(row, pad=5)
        row += 1

        label = tk.Label(l_frame, text='Depth:')
        label.grid(row=row, column=0, padx=(0,10), pady=(20,0), sticky=tk.W)
        self.depth_var = tk.StringVar(value=3)
        depth_entry = tk.Entry(l_frame, textvariable=self.depth_var, width=5,
            justify=tk.RIGHT)
        depth_entry.grid(row=row, column=1, sticky=tk.W, pady=(20,0))
        l_frame.rowconfigure(row, pad=5)
        row += 1

        button = tk.Button(l_frame, text='Draw', width=12, command=self.draw)
        button.grid(row=row, column=0, columnspan=2)
        row += 1

        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=10, pady=(0, 10),
            side=tk.RIGHT, fill=tk.BOTH, expand=True)

        # If the user presses Enter, draw.
        widgets = [
            axiom_entry,
            ignore_entry,
            dtheta_entry,
            theta0_entry,
            scale_entry,
            depth_entry,
        ]
        for widget in widgets:
            widget.bind('<Return>', lambda event: self.draw())

        # When the user selects a combo entry, load that fractal's parameters.
        self.curve_combo.bind("<<ComboboxSelected>>", self.combo_selected)
        self.select_fractal(self.curve_combo.get())

        # Display the window.
        axiom_entry.focus_force()
        self.window.mainloop()

    def combo_selected(self, event):
        '''Display the selected fractal's parameters.'''
        self.select_fractal(event.widget.get())

    def select_fractal(self, fractal_name):
        '''Display the selected fractal's parameters.'''
        lsystem = self.lsystems[fractal_name]
        rules, axiom, ignore, dtheta, max_depth, theta0, scale, \
            nice_depth = lsystem
        if False: # Display the values for debugging.
            print(lsystem)
            print(f'{axiom=}')
            print(f'{ignore=}')
            print(f'{rules=}')
            print(f'{dtheta=}')
            print(f'{theta0=}')
            print(f'{scale=}')
            print(f'{nice_depth=}')
        self.axiom_var.set(axiom)
        self.ignore_var.set(ignore)
        self.rules_text.delete('1.0', tk.END)
        for key, value in rules.items():
            self.rules_text.insert(tk.END, f'{key}: {value}\n')
        self.dtheta_var.set(dtheta)
        self.theta0_var.set(theta0)
        self.scale_var.set(scale)
        self.depth_var.set(nice_depth)

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def draw(self):
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Get the fractal's parameters.
        axiom = self.axiom_var.get()
        ignore = self.ignore_var.get()
        rules_text = self.rules_text.get('1.0', tk.END)
        dtheta = get_int('dtheta', self.dtheta_var, -360, 360)
        if dtheta is None: return
        theta0 = get_int('theta0', self.theta0_var, -360, 360)
        if theta0 is None: return
        scale = get_float('scale', self.scale_var, -10, 10)
        if scale is None: return
        depth = get_int('depth', self.depth_var, 0, 100)
        if depth is None: return

        # Convert anles to radians.
        dtheta = math.radians(dtheta)
        theta0 = math.radians(theta0)

        # Turn the rules text into a dictionary.
        rules = {}
        for line in rules_text.splitlines():
            line = line.strip()
            if len(line) == 0: continue
            key = line.split(':')[0].strip()
            value = line.split(':')[1].strip()
            rules[key] = value

        # See if it's the Sierpiński Arrowhead.
        if self.curve_combo.get() == 'Sierpiński Arrowhead':
            # Draw odd-numbered depths backward so they all point down.
            if depth % 2 == 1:
                theta0 = math.radians(180)

        if False: # Display the values for debugging.
            print(f'{axiom=}')
            print(f'{ignore=}')
            print(f'{rules=}')
            print(f'{dtheta=}')
            print(f'{theta0=}')
            print(f'{scale=}')
            print(f'{depth=}')

        # Generate the curve.
        paths = get_l_system_paths(depth, axiom, ignore, rules, dtheta,
            theta0, scale)

        # Draw the path(s).
        num_points = 0
        curves = []
        for path in paths:
            # print(path)
            curves.append(self.canvas.create_line(path, width=0, fill='black'))
            num_points += len(path)

        # Scale to fit.
        fit_curves_to_canvas(self.canvas, curves, 10)

        # Print the number of pathcs and points.
        print(f'{len(paths):,} paths, {num_points:,} points')

# In[19]:
LSystemsApp()
print('Done')
